<?php

declare(strict_types=1);

namespace Imi\Model;

use Imi\Bean\BeanFactory;
use Imi\Db\Query\Result;
use Imi\Model\Event\ModelEvents;
use Imi\Model\Event\Param\AfterQueryEventParam;

/**
 * 模型查询结果集类.
 */
class ModelQueryResult extends Result
{
    /**
     * 是否设置模型序列化字段.
     */
    protected bool $isSetSerializedFields = false;

    /**
     * 关联查询预加载字段.
     */
    protected ?array $with = null;

    /**
     * 关联查询预加载字段.
     */
    protected ?array $withField = null;

    /**
     * {@inheritDoc}
     */
    public function get(?string $className = null)
    {
        if (!$this->isSuccess)
        {
            throw new \RuntimeException('Result is not success!');
        }
        $record = $this->statementRecords[0] ?? null;
        if (!$record)
        {
            return null;
        }

        if (null === $className)
        {
            $className = $this->modelClass;
            if (null === $className)
            {
                return $record;
            }
            $isModel = true;
        }
        else
        {
            $isModel = false;
        }
        if ($isModel || is_subclass_of($className, Model::class))
        {
            $with = $this->with;
            $withField = $this->withField;
            /** @var Meta $meta */
            $meta = $className::__getMeta();
            if ($with)
            {
                if ($withField)
                {
                    $serializedFields = $withField;
                }
                else
                {
                    if ($this->isSetSerializedFields)
                    {
                        $serializedFields = $this->parseFieldNames($meta->getSerializableFieldNames(), array_keys($record));
                    }
                    else
                    {
                        $serializedFields = $meta->getParsedSerializableFieldNames();
                    }
                    foreach ($with as $k => $v)
                    {
                        if (\is_string($k))
                        {
                            $serializedFields[] = $k;
                        }
                        else
                        {
                            $serializedFields[] = $v;
                        }
                    }
                }
                /** @var Model $object */
                $object = $className::createFromRecord($record, false);
                $object->__setSerializedFields($serializedFields);
                if ($meta->hasRelation())
                {
                    ModelRelationManager::initModels([$object], null, $with, $className);
                }
            }
            else
            {
                /** @var Model $object */
                $object = $className::createFromRecord($record);
                if ($withField)
                {
                    $object->__setSerializedFields($withField);
                }
                elseif ($this->isSetSerializedFields)
                {
                    $object->__setSerializedFields($this->parseFieldNames($meta->getSerializableFieldNames(), array_keys($record)));
                }
            }
            if ($meta->isBean())
            {
                $object->trigger(ModelEvents::AFTER_QUERY, [
                    'model'      => $object,
                ], $object, AfterQueryEventParam::class);
            }
        }
        else
        {
            $object = BeanFactory::newInstance($className);
            foreach ($record as $k => $v)
            {
                $object->{$k} = $v;
            }
        }

        return $object;
    }

    /**
     * {@inheritDoc}
     */
    public function getArray(?string $className = null): array
    {
        if (!$this->isSuccess)
        {
            throw new \RuntimeException('Result is not success!');
        }

        $statementRecords = $this->statementRecords;
        if (!$statementRecords)
        {
            return [];
        }
        if (null === $className)
        {
            $className = $this->modelClass;
            if (null === $className)
            {
                return $statementRecords;
            }
            $isModel = true;
        }
        else
        {
            $isModel = false;
        }
        $list = [];
        if ($isModel || is_subclass_of($className, Model::class))
        {
            /** @var Meta $meta */
            $meta = $className::__getMeta();
            $hasRelation = $meta->hasRelation();
            $isBean = $meta->isBean();
            $withField = $this->withField;
            $with = $this->with;
            if ($with)
            {
                if ($withField)
                {
                    $serializedFields = $withField;
                }
                else
                {
                    if ($this->isSetSerializedFields)
                    {
                        $serializedFields = $this->parseFieldNames($meta->getSerializableFieldNames(), array_keys($statementRecords[0]));
                    }
                    else
                    {
                        $serializedFields = $meta->getParsedSerializableFieldNames();
                    }
                    foreach ($with as $k => $v)
                    {
                        if (\is_string($k))
                        {
                            $serializedFields[] = $k;
                        }
                        else
                        {
                            $serializedFields[] = $v;
                        }
                    }
                }
            }
            elseif ($withField)
            {
                $serializedFields = $withField;
            }
            elseif ($this->isSetSerializedFields)
            {
                $serializedFields = $this->parseFieldNames($meta->getSerializableFieldNames(), array_keys($statementRecords[0]));
            }
            else
            {
                $serializedFields = [];
            }
            foreach ($statementRecords as $item)
            {
                $list[] = $object = $className::createFromRecord($item, false);
                if ($serializedFields)
                {
                    $object->__setSerializedFields($serializedFields);
                }
                if ($isBean && !$hasRelation)
                {
                    $object->trigger(ModelEvents::AFTER_QUERY, [
                        'model' => $object,
                    ], $object, AfterQueryEventParam::class);
                }
            }
            if ($hasRelation)
            {
                ModelRelationManager::initModels($list, null, $with, $className);
                if ($isBean)
                {
                    foreach ($list as $object)
                    {
                        $object->trigger(ModelEvents::AFTER_QUERY, [
                            'model' => $object,
                        ], $object, AfterQueryEventParam::class);
                    }
                }
            }
        }
        else
        {
            foreach ($this->statementRecords as $item)
            {
                $list[] = $row = BeanFactory::newInstance($className, $item);
                foreach ($item as $k => $v)
                {
                    $row->{$k} = $v;
                }
            }
        }

        return $list;
    }

    /**
     * Get 是否设置模型序列化字段.
     */
    public function getIsSetSerializedFields(): bool
    {
        return $this->isSetSerializedFields;
    }

    /**
     * Set 是否设置模型序列化字段.
     *
     * @param bool $isSetSerializedFields 是否设置模型序列化字段
     */
    public function setIsSetSerializedFields(bool $isSetSerializedFields): self
    {
        $this->isSetSerializedFields = $isSetSerializedFields;

        return $this;
    }

    /**
     * Get 关联查询预加载字段.
     */
    public function getWith(): ?array
    {
        return $this->with;
    }

    /**
     * Set 关联查询预加载字段.
     */
    public function setWith(?array $with): self
    {
        $this->with = $with;

        return $this;
    }

    /**
     * Get 关联查询预加载字段.
     */
    public function getWithField(): ?array
    {
        return $this->withField;
    }

    /**
     * Set 关联查询预加载字段.
     */
    public function setWithField(?array $withField): self
    {
        $this->withField = $withField;

        return $this;
    }

    private function parseFieldNames(array $serializableFieldNames, array $fieldNames): array
    {
        foreach ($fieldNames as &$name)
        {
            // @deprecated 3.0 将废弃动态字段名
            if (!isset($serializableFieldNames[$name]) && false !== ($index = array_search($name, $serializableFieldNames)))
            {
                $name = $index;
            }
        }

        return $fieldNames;
    }
}
